from functools import lru_cache


class FunctionDispatch(object):
    """
    FunctionDispatch is similar to functools.singledispatch, but
    instead dispatches based on functions that take the type of the
    first argument in the method, and return True or False.

    objects that help determine dispatch should be instantiated objects.
    """

    __slots__ = ("_handler_pairs", "dispatch")

    def __init__(self):
        self._handler_pairs = []
        self.dispatch = lru_cache(64)(self._dispatch)

    def register(self, can_handle, func):
        self._handler_pairs.insert(0, (can_handle, func))
        self.dispatch.cache_clear()

    def _dispatch(self, typ):
        """
        returns the appropriate handler, for the object passed.
        """
        for can_handle, handler in self._handler_pairs:
            # can handle could raise an exception here
            # such as issubclass being called on an instance.
            # it's easier to just ignore that case.
            try:
                if can_handle(typ):
                    return handler
            except Exception:
                pass
        raise KeyError("unable to find handler for {0}".format(typ))



from typing import Any, Callable, Dict, List, Literal, Set, Type, _GenericAlias

from cattr import Converter


def make_literal_structure_fn(
    converter: Converter, *classes: Type,
) -> Callable[[Dict, Type], Any]:
    """
    Given a sequence of attrs classes, each of which has a different Literal
    field, generate a union structuring function.
    """
    # First, find the field with the literal.
    literal_fields: List[Set[str]] = []

    for cls in classes:
        lits = set()
        for attr in cls.__attrs_attrs__:
            t = attr.type
            if isinstance(t, _GenericAlias) and t.__origin__ is Literal:
                lits.add(attr.name)
        literal_fields.append(lits)
    field = next(iter(set.intersection(*literal_fields)))

    cls_mapping = {
        getattr(cls.__attrs_attrs__, field).type.__args__[0]: cls
        for cls in classes
    }

    def union_literal_hook(obj: Dict, _: Type):
        val = obj[field]
        try:
            cls = cls_mapping[val]
        except KeyError:
            raise Exception(f"No class matches literal {val}.")
        return converter.structure(obj, cls)

    return union_literal_hook



from functools import lru_cache, singledispatch

import attr

from .function_dispatch import FunctionDispatch


@attr.s
class _DispatchNotFound(object):
    """ a dummy object to help signify a dispatch not found """

    pass


class MultiStrategyDispatch(object):
    """
    MultiStrategyDispatch uses a
    combination of FunctionDispatch and singledispatch.

    singledispatch is attempted first. If nothing is
    registered for singledispatch, or an exception occurs,
    the FunctionDispatch instance is then used.
    """

    __slots__ = ("_function_dispatch", "_single_dispatch", "dispatch")

    def __init__(self, fallback_func):
        self._function_dispatch = FunctionDispatch()
        self._function_dispatch.register(lambda cls: True, fallback_func)
        self._single_dispatch = singledispatch(_DispatchNotFound)
        self.dispatch = lru_cache(maxsize=None)(self._dispatch)

    def _dispatch(self, cl):
        try:
            dispatch = self._single_dispatch.dispatch(cl)
            if dispatch is not _DispatchNotFound:
                return dispatch
        except Exception:
            pass
        return self._function_dispatch.dispatch(cl)

    def register_cls_list(self, cls_and_handler):
        """ register a class to singledispatch """
        for cls, handler in cls_and_handler:
            self._single_dispatch.register(cls, handler)
        self.dispatch.cache_clear()

    def register_func_list(self, func_and_handler):
        """register a function to determine if the handle
        should be used for the type
        """
        for func, handler in func_and_handler:
            self._function_dispatch.register(func, handler)
        self.dispatch.cache_clear()
